"""
Elapsed time timer.
"""

from __future__ import absolute_import

import datetime


class ElapsedTimer(object):
    """
    Elapsed time timer.

    Calculates elapsed time between initiation/reset and access.
    """

    def __init__(self):
        """ Initiate the object with current time"""
        self.start_time = datetime.datetime.now()

    def reset(self):
        """ Reset the start time for the timer"""
        self.start_time = datetime.datetime.now()

    def elapsed_ms(self):
        """ Get the elapsed time in milliseconds. returns floating
            point representation of elapsed time in seconds.
        """
        dt = datetime.datetime.now() - self.start_time
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000 + \
            dt.microseconds / 1000.0

    def elapsed_sec(self):
        """ get the elapsed time in seconds. Returns floating
            point representation of time in seconds
        """
        return self.elapsed_ms() / 1000



#!/usr/bin/env python
#

"""
Unittest functions in _cim_operations.
"""

from __future__ import print_function, absolute_import

import sys
import os
import io
import re
import pytest

from ...utils import skip_if_moftab_regenerated
from ..utils.dmtf_mof_schema_def import install_test_dmtf_schema
from ..utils.pytest_extensions import simplified_test_function
from ..utils.unittest_extensions import assert_copy

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMConnection, ParseError, DEFAULT_NAMESPACE, \
    CIMError, CIMInstanceName, CIMClassName, CIMInstance, \
    CIMClass, DEFAULT_TIMEOUT  # noqa: E402
# pylint: disable=redefined-builtin
from pywbem import ConnectionError  # noqa: E402
# pylint: enable=redefined-builtin
from pywbem._recorder import LogOperationRecorder  # noqa: E402
from pywbem._recorder import TestClientRecorder as \
    MyTestClientRecorder  # noqa: E402
pywbem_mock = import_installed('pywbem_mock')
from pywbem_mock import FakedWBEMConnection  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# Literal form {"blah: 0} faster than dict(blah=0) but same functionality
# pylint: disable=use-dict-literal

# Name of null device
DEV_NULL = 'nul' if sys.platform == 'win32' else '/dev/null'

OK = True     # mark tests OK when they execute correctly
RUN = True    # Mark OK = False and current test case being created RUN
FAIL = False  # Any test currently FAILING or not tested yet
SKIP = False


TESTCASES_INIT_WBEMCONNECTION = [

    # Test WBEMConnection init
    # The testcase items are tuples with these items:
    # * desc: Testcase description
    # * files: empty files to be created before the test
    # * init_kwargs: dict of init arguments for WBEMConnection, in addition to
    #   the url argument
    # * exp_attrs: dict of expected WBEMConnection attributes after init
    # * exp_exc: Type of expected exception, or None
    # * exp_exc_regex: Regexp for the expected exception message, or None

    (
        "No optional init parameters, test defaults",
        [],
        {},
        dict(
            creds=None,
            default_namespace='root/cimv2',
            x509=None,
            ca_certs=None,
            no_verification=False,
            timeout=DEFAULT_TIMEOUT,
            use_pull_operations=False,
            stats_enabled=False,
            proxies=None,
        ),
        None, None
    ),
    (
        "All optional init parameters, except x509 and ca_certs",
        [],
        dict(
            creds=('myuser', 'mypw'),
            default_namespace='root/myns',
            no_verification=True,
            timeout=20,
            use_pull_operations=True,
            stats_enabled=True,
            proxies=None,
        ),
        dict(
            creds=('myuser', 'mypw'),
            default_namespace='root/myns',
            no_verification=True,
            timeout=20,
            use_pull_operations=True,
            stats_enabled=True,
            proxies=None,
        ),
        None, None
    ),
    (
        "x509 parameter that is a dict with existing cert_file and "
        "existing key_file",
        ['mycertfile.tmp', 'mykeyfile.tmp'],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file='mykeyfile.tmp',
            ),
        ),
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file='mykeyfile.tmp',
            ),
        ),
        None, None
    ),
    (
        "x509 parameter that is a dict with existing cert_file and "
        "omitted key_file",
        ['mycertfile.tmp'],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
            ),
        ),
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
            ),
        ),
        None, None
    ),
    (
        "x509 parameter that is a dict with non-existing cert_file "
        "(invalid) and existing key_file",
        ['mykeyfile.tmp'],
        dict(
            x509=dict(
                cert_file='nonexistingcertfile.tmp',
                key_file='mykeyfile.tmp',
            ),
        ),
        {},
        IOError, "Client certificate file .* not found"
    ),
    (
        "x509 parameter that is a dict with existing cert_file "
        "and non-existing key_file (invalid)",
        ['mycertfile.tmp'],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file='nonexistingkeyfile.tmp',
            ),
        ),
        {},
        IOError, "Client key file .* not found"
    ),
    (
        "x509 parameter that is a dict with cert_file=None (invalid) "
        "and key_file=None",
        [],
        dict(
            x509=dict(
                cert_file=None,
                key_file=None,
            ),
        ),
        {},
        TypeError, "cert_file.* must be a string"
    ),
    (
        "x509 parameter that is a dict with cert_file=3 (invalid) "
        " Must be a string",
        [],
        dict(
            x509=dict(
                cert_file=3,
                key_file=None,
            ),
        ),
        {},
        TypeError, "The 'cert_file' item in the x509 parameter must be a string"
    ),
    (
        "x509 parameter that is a dict with cert_file='mycertfile.tmp' "
        " keyfile invalid type",
        [],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file=3,
            ),
        ),
        {},
        TypeError, "The 'key_file' item in the x509 parameter must be a string",
    ),
    (
        "x509 parameter that is a dict with no cert_file key "
        "and key_file=None",
        [],
        dict(
            x509=dict(
                key_file=None,
            ),
        ),
        {},
        ValueError, "The x509 parameter does not have the required key "
        "'cert_file'"
    ),

    (
        "x509 parameter that is an existing file (invalid)",
        ['mycertfile.tmp'],
        dict(
            x509='mycertfile.tmp',
        ),
        {},
        TypeError, "x509 .* must be a dictionary"
    ),
    (
        "ca_certs parameter that is an existing file",
        ['mycacertfile.tmp'],
        dict(
            ca_certs='mycacertfile.tmp',
        ),
        dict(
            ca_certs='mycacertfile.tmp',
        ),
        None, None
    ),
    (
        "ca_certs parameter that is an integer (invalid)",
        [],
        dict(
            ca_certs=42,
        ),
        {},
        TypeError, "ca_certs .* invalid type"
    ),
    (
        "ca_certs parameter that is a non-existing file (invalid)",
        [],
        dict(
            ca_certs='mycacertfile.tmp',
        ),
        {},
        IOError, "file or directory not found"
    ),
    (
        "proxies parameter that is valid",
        [],
        dict(
            proxies={
                'http': 'http://user:pass@10.10.1.10:3128',
                'https': 'http://user:pass@10.10.1.10:1080',
            },
        ),
        dict(
            proxies={
                'http': 'http://user:pass@10.10.1.10:3128',
                'https': 'http://user:pass@10.10.1.10:1080',
            },
        ),
        None, None
    ),
    (
        "proxies parameter that is an invalid type",
        [],
        dict(
            proxies=42,
        ),
        {},
        TypeError, "proxies .* must be a dictionary .*"
    ),
]


class TestCreateConnection(object):
    """
    Test construction of WBEMConnection and those functions that do not
    depend on actually creating a connection
    """

    @pytest.mark.parametrize(
        'desc, files, init_kwargs, exp_attrs, exp_exc, exp_exc_regex',
        TESTCASES_INIT_WBEMCONNECTION)
    def test_init(self, desc, files, init_kwargs, exp_attrs, exp_exc,
                  exp_exc_regex):
        # pylint: disable=no-self-use,unused-argument
        """
        Test initialization of a WBEMConnection object.
        """

        try:
            for fname in files:
                # create empty file
                with io.open(fname, 'a', encoding='utf-8'):
                    pass

            if exp_exc is not None:
                with pytest.raises(exp_exc) as exc_info:

                    # The code to be tested
                    WBEMConnection('http://localhost', **init_kwargs)

                exc = exc_info.value
                exc_msg = str(exc)
                if exp_exc_regex:
                    assert re.search(exp_exc_regex, exc_msg)
            else:

                # The code to be tested
                conn = WBEMConnection('http://localhost', **init_kwargs)

                for name in exp_attrs:
                    exp_value = exp_attrs[name]
                    act_value = getattr(conn, name)
                    assert act_value == exp_value

        finally:
            for fname in files:
                os.remove(fname)

    @pytest.mark.parametrize(
        'attr_name, value', [
            ('url', 'http://myserver:5988'),
            ('creds', ('x', 'y')),
            ('x509', dict(cert_file='c', key_file='k')),
            ('ca_certs', 'xxx'),
            ('no_verification', True),
            ('last_raw_request', '<CIM/>'),
            ('last_raw_reply', '<CIM/>'),
            ('last_raw_reply', '<CIM/>'),
            ('last_request', '<CIM/>'),
            ('last_request_len', 7),
            ('last_reply_len', 7),
        ])
    def test_attrs_readonly(self, attr_name, value):
        # pylint: disable=no-self-use
        """
        Test that certain attributes that were previously settable are now
        read-only and that the original value has not changed when trying to
        modify them.
        """
        conn = WBEMConnection('http://localhost')
        value1 = getattr(conn, attr_name)
        with pytest.raises(AttributeError):

            # The code to be tested
            setattr(conn, attr_name, value)

        value2 = getattr(conn, attr_name)
        assert value2 == value1

    @pytest.mark.parametrize(
        'kwargs, exp_default_namespace', [
            ({}, DEFAULT_NAMESPACE),
            (dict(default_namespace=None), DEFAULT_NAMESPACE),
            (dict(default_namespace=DEFAULT_NAMESPACE), DEFAULT_NAMESPACE),
            (dict(default_namespace='blah'), 'blah'),
        ])
    def test_init_default_namespace(self, kwargs, exp_default_namespace):
        # pylint: disable=no-self-use
        """Test creation of wbem connection with default_namespace values"""
        conn = WBEMConnection('http://localhost', ('name', 'pw'), **kwargs)
        assert conn.default_namespace == exp_default_namespace

    def test_namespace_slashes_init(self):  # pylint: disable=no-self-use
        """Test stripping of leading and trailing slashes in default namespace
        of wbem connection when initializing"""
        conn = WBEMConnection('http://localhost', None,
                              default_namespace='//root/blah//')
        assert conn.default_namespace == 'root/blah'

    def test_namespace_slashes_set(self):  # pylint: disable=no-self-use
        """Test stripping of leading and trailing slashes in default namespace
        of wbem connection when setting the attribute"""
        conn = WBEMConnection('http://localhost', None,
                              default_namespace=None)
        conn.default_namespace = '//root/blah//'
        assert conn.default_namespace == 'root/blah'

    def test_add_operation_recorder(self):  # pylint: disable=no-self-use
        """Test addition of an operation recorder"""
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
        # pylint: disable=protected-access
        assert len(conn._operation_recorders) == 1

    def test_add_operation_recorders(self):  # pylint: disable=no-self-use
        """Test addition of multiple operation recorders"""
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
        tcr_fn = 'blah.yaml'
        tcr_file = MyTestClientRecorder.open_file(tcr_fn, 'a')
        conn.add_operation_recorder(MyTestClientRecorder(tcr_file))
        # pylint: disable=protected-access
        assert len(conn._operation_recorders) == 2
        tcr_file.close()
        os.remove(tcr_fn)

    def test_add_same_twice(self):  # pylint: disable=no-self-use
        """
        Test addition of same recorder twice. This not allowed so
        generates exception
        """
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
        # pylint: disable=protected-access
        assert len(conn._operation_recorders) == 1

        with pytest.raises(ValueError):
            conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))

    def test_stats_enabled(self):  # pylint: disable=no-self-use
        """
        Test the property stats_enabled setter
        """
        conn = WBEMConnection('http://localhost')
        assert conn.stats_enabled is False
        conn.stats_enabled = True
        assert conn.stats_enabled is True
        conn.stats_enabled = False
        assert conn.stats_enabled is False

    def test_operation_recorder_enabled(self):  # pylint: disable=no-self-use
        """
        Test the property operation_recorder enabled setter
        """
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
        assert conn.operation_recorder_enabled is True
        conn.operation_recorder_enabled = False
        assert conn.operation_recorder_enabled is False
        conn.operation_recorder_enabled = True
        assert conn.operation_recorder_enabled is True

    def test_close(self):  # pylint: disable=no-self-use
        """
        Test closing a connection once.
        """
        conn = WBEMConnection('http://localhost')
        assert conn.session is not None
        conn.close()
        assert conn.session is None

    def test_close_twice(self):  # pylint: disable=no-self-use
        """
        Test closing a connection twice.
        """
        conn = WBEMConnection('http://localhost')
        conn.close()
        assert conn.session is None
        with pytest.raises(ConnectionError) as exc_info:
            conn.close()
            exc = exc_info.value
            assert re.match(r'is closed', str(exc))

    def test_close_imethod(self):  # pylint: disable=no-self-use
        """
        Test closing a connection and invoking an intrinsic operation.
        """
        conn = WBEMConnection('http://localhost')
        conn.close()
        assert conn.session is None
        with pytest.raises(ConnectionError) as exc_info:
            conn.GetClass('CIM_Foo')
            exc = exc_info.value
            assert re.match(r'is closed', str(exc))

    def test_close_method(self):  # pylint: disable=no-self-use
        """
        Test closing a connection and invoking a CIM method.
        """
        conn = WBEMConnection('http://localhost')
        conn.close()
        assert conn.session is None
        with pytest.raises(ConnectionError) as exc_info:
            conn.InvokeMethod(ObjectName='CIM_Foo', MethodName='M1')
            exc = exc_info.value
            assert re.match(r'is closed', str(exc))

    def test_close_export(self):  # pylint: disable=no-self-use
        """
        Test closing a connection and invoking an export operation.
        """
        conn = WBEMConnection('http://localhost')
        conn.close()
        assert conn.session is None
        indi = CIMInstance("CIM_AlertIndication")
        with pytest.raises(ConnectionError) as exc_info:
            conn.ExportIndication(NewIndication=indi)
            exc = exc_info.value
            assert re.match(r'is closed', str(exc))

    def test_context_manager(self):  # pylint: disable=no-self-use
        """
        Test the use of WBEMConnection as a context manager.
        """
        with WBEMConnection('http://localhost:5988') as conn:
            assert isinstance(conn, WBEMConnection)
            assert conn.url == 'http://localhost:5988'
            assert conn.session is not None
        assert conn.session is None


def test_conn_set_timeout():
    """
    Test setting the 'timeout' property of WBEMConnection.
    """
    conn = WBEMConnection('http://localhost', timeout=20)
    assert conn.timeout == 20
    conn.timeout = 25
    assert conn.timeout == 25


class TestGetRsltParams(object):
    """Test WBEMConnection._get_rslt_params method."""

    @pytest.mark.parametrize(
        'irval', [None, 'bla', [], {}])
    @pytest.mark.parametrize(
        # enumctxt input, eos input, expected eos, expected exception type
        'ec, eos, eos_exp, exctype_exp', [
            # following are successul returns
            [u'contextblah', u'False', False, None],  # normal enumctxt rtn
            ["", u'True', True, None],                # eos true, enmctxt empty
            [None, u'True', True, None],              # eos true, enmctxt None
            [u'contextblah', u'True', True, None],    # eos tru, cts with value
            # following are exceptions
            [None, None, None, ParseError],      # no values in eos or enumctxt
            [None, u'False', None, ParseError],  # no value in ec and eos False
        ]
    )
    def test_with_params(self, irval, ec, eos, eos_exp, exctype_exp):
        # pylint: disable=no-self-use,protected-access
        """
        Test combminations of IRETURNVALUE, EOS and EnumerationContext for
        both good and fail responses.
        """
        conn = WBEMConnection('http://localhost')
        result = [
            (u'IRETURNVALUE', {}, irval),
            (u'EnumerationContext', None, ec),
            (u'EndOfSequence', None, eos)
        ]

        if exctype_exp:
            with pytest.raises(exctype_exp):

                # pylint: disable=protected-access
                result = conn._get_rslt_params(result, 'root/blah')

        else:

            # pylint: disable=protected-access
            result = conn._get_rslt_params(result, 'root/blah')

            # the _get_rslt_params method sets context to None if eos True
            ecs = None if eos_exp is True else (ec, 'root/blah')
            assert result == (irval, eos_exp, ecs)

    def test_missing(self):
        # pylint: disable=no-self-use
        """
        Test both Enum context and EndOfSequence completely missing.
        Generates exception
        """
        conn = WBEMConnection('http://localhost')
        result = [
            (u'IRETURNVALUE', {}, {})
        ]

        with pytest.raises(ParseError):

            # pylint: disable=protected-access
            result = conn._get_rslt_params(result, 'namespace')


def build_repo():
    """Fixture to initialize the mock environment and install classes.
       Defined as a function without the pytest.fixture decorator because
       there is no way to include a fixture in the signature of a test
       function that uses simplified_test_function. Simplified_test_function
       reports a parameter count difference because of the extram parameter
       that is the fixture and fixtures may not be included in the testcases.
    """
    schema = install_test_dmtf_schema()
    namespace = "root/cimv2"
    partial_schema = """
        #pragma locale ("en_US")
        #pragma include ("Interop/CIM_ObjectManager.mof")
        #pragma include ("Interop/CIM_RegisteredProfile.mof")
        """

    conn = FakedWBEMConnection(default_namespace=namespace)
    conn.compile_mof_string(partial_schema, namespace=namespace,
                            search_paths=[schema.schema_mof_dir])
    return conn


TESTCASES_IS_SUBCLASS = [

    # Testcases for WBEMConnection.is_subclass()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to is_subclass().
    #   * init_kwargs: Dict of keyword arguments to is_subclass().
    #   * exp_attrs: Dict of expected attributes of resulting object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger

    (
        "Test of valid call with strings that returns True",
        dict(
            init_args=['CIM_ObjectManager', 'CIM_ManagedElement'],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call with CIM objects that returns True",
        dict(
            init_args=[
                CIMClass('CIM_ObjectManager', superclass='CIM_WBEMService'),
                CIMClass('CIM_ManagedElement', superclass=None),
            ],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call with CIM object and string that returns True",
        dict(
            init_args=[
                CIMClass('CIM_ObjectManager', superclass='CIM_WBEMService'),
                'CIM_ManagedElement',
            ],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call with string and CIM object that returns True",
        dict(
            init_args=[
                'CIM_ObjectManager',
                CIMClass('CIM_ManagedElement', superclass=None),
            ],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call that returns False",
        dict(
            init_args=['CIM_ManagedElement', 'CIM_ObjectManager'],
            exp_attrs=False,
        ),
        None, None, True
    ),
    (
        "Test of class equal to superclass",
        dict(
            init_args=['CIM_ObjectManager', 'CIM_ObjectManager'],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of superclass not in repo, returns False",
        dict(
            init_args=['CIM_ManagedElement', 'CIM_Blah'],
            exp_attrs=False,
        ),
        None, None, True
    ),
    (
        "Test of class not in repo (error)",
        dict(
            init_args=['CIM_Blah', 'CIM_ManagedElement'],
            exp_attrs=None,
        ),
        CIMError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_IS_SUBCLASS)
@simplified_test_function
def test_is_subclass(testcase, init_args, exp_attrs):
    # pylint: disable=unused-argument
    """
    Test the method WBEMConnection.is_subclass().
    """

    skip_if_moftab_regenerated()

    conn = build_repo()   # we wanted this to be fixture.
    # The code to be tested
    result = conn.is_subclass(conn.default_namespace, *init_args)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None
    assert exp_attrs == result


# TODO List of failure tests to be tested with following test
# methodcall. test objectname
# Test Iter invalid maxcount >- 0 and OperationTimeout < 0
# line 2008, methodcall request with classname no namespace  - not a failue case
# line 2010, methodcall request with classname, use string  - not a failure case
# request with object name not CIMInstanceName, CIMClassName, not string - DONE
# test _iparam_objectname type error if not CIMClassname, instancename,
#     string - DONE
# iparam_classname not CIMClassName or string - DONE
# iparam_instancename not CIMInstanceName or None  - DONE
# modified instance. no path,
# Modified instance no class in path
# modified_instance no classname in modifiedinstance line 3022
# line 4595, iterenumerateinstances CIMClassName input no namespace
# line 4612, close iterenum early.  Not sure this can be done.

TESTCASES_REQUEST_INVALID_PARAMS = [

    # Testcases for WBEMConnection  method calls with invalid params.

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_kwargs: keyword args for WBEMConnection __init
    #   * method: operation method to call
    #   * args=: Arguments for the operation method being called
    #   * kwargs: Dict of keyword arguments for the WBEMConnection __init
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger
    (
        "Test invalid invokemethod objectname (integer)",
        dict(
            init_kwargs={},
            method='invokemethod',
            args=["MethodName", 1],
            kwargs={},
        ),
        TypeError, None, OK
    ),
    (
        "Test invalid invokemethod objectname is None",
        dict(
            init_kwargs={},
            method='invokemethod',
            args=["MethodName", None],
            kwargs={},
        ),
        TypeError, None, OK
    ),
    (
        "Test invalid invokemethod inferred object integer None",
        dict(
            init_kwargs={},
            method='invokemethod',
            args=[None, "ObjectName"],
            kwargs={"param1": 1},
        ),
        TypeError, None, False  # Fails with connection failure.
    ),
    (
        "Test enumerateinstances, fails, classname is not CIMClassname or Str",
        dict(
            init_kwargs={},
            method='enumerateinstances',
            args=[CIMInstanceName("CIM_Blah")],
            kwargs={},
        ),
        TypeError, None, OK
    ),
    (
        "Test GetInstance, invalid InstanceName tyhpe",
        dict(
            init_kwargs={},
            method='getinstance',
            args=[CIMClassName("CIM_Blah")],
            kwargs={},
        ),
        TypeError, None, OK
    ),

    (
        "Test ModifyInstance, No path in ModifiedInstance",
        dict(
            init_kwargs={},
            method='modifyinstance',
            args=[CIMInstance("CIMBlah")],
            kwargs={},
        ),
        ValueError, None, OK
    ),

    (
        "Test EnumerateInstances, Invalid args type",
        dict(
            init_kwargs={},
            method='enumerateinstances',
            args=[CIMClass("CIMBlah")],
            kwargs={},
        ),
        TypeError, None, OK
    ),

    (
        "Test OpenEnumerateInstances, Invalid type for operationtimeout",
        dict(
            init_kwargs={},
            method='openenumerateinstances',
            args=[CIMClass("CIMBlah")],
            kwargs={"OperationTimeout": "shouldbeinteger"},
        ),
        TypeError, None, OK
    ),

    (
        "Test OpenEnumerateInstances, Invalid value for operationtimeout",
        dict(
            init_kwargs={},
            method='openenumerateinstances',
            args=[CIMClass("CIMBlah")],
            kwargs={"OperationTimeout": -30},
        ),
        TypeError, None, OK
    ),

    (
        "Test OpenEnumerateInstancepaths, Invalid type for operationtimeout",
        dict(
            init_kwargs={},
            method='openenumerateinstancepaths',
            args=[CIMClass("CIMBlah")],
            kwargs={"OperationTimeout": "shouldbeinteger"},
        ),
        TypeError, None, OK
    ),

    (
        "Test OpenEnumerateInstancePaths, Invalid value for operationtimeout",
        dict(
            init_kwargs={},
            method='openenumerateinstancepaths',
            args=[CIMClass("CIMBlah")],
            kwargs={"OperationTimeout": -30},
        ),
        TypeError, None, OK
    ),

    (
        "Test pullinstancewithpath, Invalid context type",
        dict(
            init_kwargs={},
            method='pullinstanceswithpath',
            args=[1, 0],
            kwargs={},
        ),
        TypeError, None, OK
    ),

    (
        "Test pullinstancewithpath, context list len ",
        dict(
            init_kwargs={},
            method='pullinstanceswithpath',
            args=[[], 0],
            kwargs={},
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, invalid timeout.lt 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(OperationTimeout=-1),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, invalid maxobjectcount == 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(MaxObjectCount=0),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, Pull not allowed. with Filterquery",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQuery="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, Pull not allowed. with FilterQueryLang",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQueryLanguage="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, invalid timeout.lt 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(OperationTimeout=-1),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, invalid maxobjectcount == 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(MaxObjectCount=0),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, Pull not allowed. with Filterquery",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQuery="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, Pull not allowed. with "
        "FilterQueryLang",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQueryLanguage="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, Pull not allowed. with "
        "ContinueOnError True",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(ContinueOnError=True),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, Pull not allowed. with "
        "ContinueOnError True",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(ContinueOnError=True),
        ),
        ValueError, None, OK
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_REQUEST_INVALID_PARAMS)
@simplified_test_function
def test_request_invalid_params(testcase, init_kwargs, method, args, kwargs):
    """Execute the defined method expecting exception"""

    conn = WBEMConnection('http://localhost', **init_kwargs)

    if method == 'invokemethod':
        _ = conn.InvokeMethod(*args, **kwargs)

    elif method == 'enumerateinstances':
        _ = conn.EnumerateInstances(*args, **kwargs)

    elif method == 'getinstance':
        _ = conn.GetInstance(*args, **kwargs)

    elif method == 'modifyinstance':
        conn.ModifyInstance(*args, **kwargs)

    elif method == 'openenumerateinstances':
        _ = conn.OpenEnumerateInstances(*args, **kwargs)

    elif method == 'openenumerateinstancepaths':
        _ = conn.OpenEnumerateInstancePaths(*args, **kwargs)

    elif method == 'pullinstanceswithpath':
        _ = conn.PullInstancesWithPath(*args, **kwargs)

    elif method == 'iterenumerateinstances':
        _ = list(conn.IterEnumerateInstances(*args, **kwargs))

    elif method == 'iterenumerateinstancepaths':
        _ = list(conn.IterEnumerateInstancePaths(*args, **kwargs))

    else:
        assert False, "Test failed. Incomplete. "  \
            "method {0} not found".format(method)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None


def test_close():
    """
    Test WBEMConnection.close().
    """

    conn = WBEMConnection('http://localhost')

    assert conn.session is not None

    # The code to be tested, good case
    conn.close()

    assert conn.session is None

    with pytest.raises(ConnectionError):

        # The code to be tested, error case
        conn.close()


TESTCASES_COPY_WBEMCONNECTION = [

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * files: empty files to be created before the test
    #   * init_kwargs: keyword args for WBEMConnection __init
    #   * set_debug: Boolean that causes debug to be enabled
    #   * add_recorder: Boolean that causes a LogRecorder to be added
    #   * enable_recorder: Boolean that causes recorders to be enabled
    #   * enable_statistics: Boolean that causes statistics to be enabled
    #   * perform_operation: Boolean that causes an operation to be performed
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger

    (
        "No optional init parameters, test defaults",
        dict(
            files=[],
            init_kwargs=dict(
                url='http://localhost:5988',
            ),
            set_debug=False,
            add_recorder=False,
            enable_recorder=False,
            enable_statistics=False,
            perform_operation=False,
        ),
        None, None, True
    ),
    (
        "All optional init parameters, except x509 and ca_certs",
        dict(
            files=[],
            init_kwargs=dict(
                url='http://localhost:5988',
                creds=('myuser', 'mypw'),
                default_namespace='root/myns',
                no_verification=True,
                timeout=20,
                use_pull_operations=True,
                stats_enabled=True,
                proxies={
                    'http': 'http://user:pass@10.10.1.10:3128',
                    'https': 'http://user:pass@10.10.1.10:1080',
                },
            ),
            set_debug=False,
            add_recorder=False,
            enable_recorder=False,
            enable_statistics=False,
            perform_operation=False,
        ),
        None, None, True
    ),
    (
        "x509 parameter that is a dict with existing cert_file and "
        "existing key_file",
        dict(
            files=['mycertfile.tmp', 'mykeyfile.tmp'],
            init_kwargs=dict(
                url='http://localhost:5988',
                x509=dict(
                    cert_file='mycertfile.tmp',
                    key_file='mykeyfile.tmp',
                ),
            ),
            set_debug=False,
            add_recorder=False,
            enable_recorder=False,
            enable_statistics=False,
            perform_operation=False,
        ),
        None, None, True
    ),
    (
        "x509 parameter that is a dict with existing cert_file and "
        "omitted key_file",
        dict(
            files=['mycertfile.tmp'],
            init_kwargs=dict(
                url='http://localhost:5988',
                x509=dict(
                    cert_file='mycertfile.tmp',
                ),
            ),
            set_debug=False,
            add_recorder=False,
            enable_recorder=False,
            enable_statistics=False,
            perform_operation=False,
        ),
        None, None, True
    ),
    (
        "with a recorder",
        dict(
            files=[],
            init_kwargs=dict(
                url='http://localhost:5988',
                creds=('myuser', 'mypw'),
                default_namespace='root/myns',
                no_verification=True,
                timeout=20,
                use_pull_operations=True,
                stats_enabled=True,
                proxies={
                    'http': 'http://user:pass@10.10.1.10:3128',
                    'https': 'http://user:pass@10.10.1.10:1080',
                },
            ),
            set_debug=True,
            add_recorder=True,
            enable_recorder=True,
            enable_statistics=True,
            perform_operation=False,
        ),
        None, None, True
    ),
    (
        "with a recorder and an operation performed",
        dict(
            files=[],
            init_kwargs=dict(
                url='http://localhost:5988',
                creds=('myuser', 'mypw'),
                default_namespace='root/myns',
                no_verification=True,
                timeout=1,
                use_pull_operations=True,
                stats_enabled=True,
                proxies={
                    'http': 'http://user:pass@10.10.1.10:3128',
                    'https': 'http://user:pass@10.10.1.10:1080',
                },
            ),
            set_debug=True,
            add_recorder=True,
            enable_recorder=True,
            enable_statistics=True,
            perform_operation=True,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_COPY_WBEMCONNECTION)
@simplified_test_function
def test_copy_conn(
        testcase, files, init_kwargs, set_debug, add_recorder, enable_recorder,
        enable_statistics, perform_operation):
    """Test WBEMConnection.copy()"""

    try:
        for fname in files:
            # create empty file
            with io.open(fname, 'a', encoding='utf-8'):
                pass

        # pylint: disable=consider-using-with
        dev_null = io.open(DEV_NULL, 'a', encoding='utf-8')

        conn = WBEMConnection(**init_kwargs)

        if set_debug:
            conn.debug = True

        if add_recorder:
            conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
            conn.add_operation_recorder(MyTestClientRecorder(dev_null))

        if enable_recorder:
            conn.operation_recorder_enabled = True

        if enable_statistics:
            conn.stats_enabled = True

        if perform_operation:
            try:
                conn.GetQualifier('Abstract')
            except pywbem.Error:  # Should time out
                pass

        # The code to be tested
        cpy = conn.copy()

        # Ensure that exceptions raised in the remainder of this function
        # are not mistaken as expected exceptions
        assert testcase.exp_exc_types is None

        # pylint: disable=protected-access,unidiomatic-type-check

        # Verify attributes that should have been copied

        assert_copy(cpy.url, conn.url)
        assert_copy(cpy.creds, conn.creds)
        assert_copy(cpy.default_namespace, conn.default_namespace)
        assert_copy(cpy.x509, conn.x509)
        assert_copy(cpy.ca_certs, conn.ca_certs)
        assert_copy(cpy.no_verification, conn.no_verification)
        assert_copy(cpy.timeout, conn.timeout)
        assert_copy(cpy.use_pull_operations, conn.use_pull_operations)
        assert_copy(cpy.stats_enabled, conn.stats_enabled)
        assert_copy(cpy.proxies, conn.proxies)
        assert_copy(cpy._scheme, conn._scheme)
        assert_copy(cpy._host, conn._host)
        assert_copy(cpy._use_enum_inst_pull_operations,
                    conn._use_enum_inst_pull_operations)
        assert_copy(cpy._use_enum_path_pull_operations,
                    conn._use_enum_path_pull_operations)
        assert_copy(cpy._use_ref_inst_pull_operations,
                    conn._use_ref_inst_pull_operations)
        assert_copy(cpy._use_ref_path_pull_operations,
                    conn._use_ref_path_pull_operations)
        assert_copy(cpy._use_assoc_inst_pull_operations,
                    conn._use_assoc_inst_pull_operations)
        assert_copy(cpy._use_assoc_path_pull_operations,
                    conn._use_assoc_path_pull_operations)
        assert_copy(cpy._use_query_pull_operations,
                    conn._use_query_pull_operations)

        # Verify attributes that should have been reset

        assert cpy._debug is False
        assert cpy._last_raw_request is None
        assert cpy._last_raw_reply is None
        assert cpy._last_request is None
        assert cpy._last_request_xml_item is None
        assert cpy._last_reply is None
        assert cpy._last_reply_xml_item is None
        assert cpy._last_request_len == 0
        assert cpy._last_reply_len == 0
        assert cpy._last_operation_time is None
        assert cpy._last_server_response_time is None

        # Session object should be a new object
        assert type(cpy.session) is type(conn.session)  # noqa: E721
        assert id(cpy.session) != id(conn.session)
        # Check that the Session object has adapters for the same schemes
        assert len(cpy.session.adapters) == len(conn.session.adapters)
        for scheme in conn.session.adapters:
            assert scheme in cpy.session.adapters
        # Beyond that, we do not determine whether the Session object is a
        # deep copy or a new object, because that would make this test too
        # much dependent on the internal structure of requests.Session.

        # Recorder objects should be new objects, but with same enabled state
        assert len(cpy._operation_recorders) == len(conn._operation_recorders)
        for i, conn_rec in enumerate(conn._operation_recorders):
            cpy_rec = cpy._operation_recorders[i]
            assert type(cpy_rec) is type(conn_rec)  # noqa: E721
            assert id(cpy_rec) != id(conn_rec)
            assert cpy_rec._pywbem_method is None
            assert cpy_rec._pywbem_args is None
            assert cpy_rec._pywbem_result_ret is None
            assert cpy_rec._pywbem_result_exc is None
            assert cpy_rec.enabled == conn_rec.enabled

        # Statistics object should be a new object, but with same enabled state
        assert type(cpy._statistics) is type(conn._statistics)  # noqa: E721
        assert id(cpy._statistics) != id(conn._statistics)
        assert not cpy._statistics._op_stats
        assert cpy._statistics.enabled == conn._statistics.enabled

        # Verify attributes that have been treated specially

        assert cpy._conn_id != conn._conn_id  # increased

        # pylint: enable=protected-access,unidiomatic-type-check

    finally:
        for fname in files:
            os.remove(fname)
        dev_null.close()


TESTCASES_WBEMCONNECTION_ERROR = [

    # Test WBEMConnection error
    # The testcase items are tuples with these items:
    # * desc: Testcase description
    # * init_kwargs: dict of init arguments for WBEMConnection
    # * exp_exc: Type of expected exception, or None
    # * exp_exc_regex: Regexp for the expected exception message, or None

    (
        "http connection",
        dict(
            url='http://InvalidHostName',
            creds=None,
            default_namespace='root/cimv2',
            x509=None,
            ca_certs=None,
            no_verification=False,
            timeout=5,
            use_pull_operations=False,
            stats_enabled=False,
            proxies=None,
        ),
        ConnectionError, None
    ),
    (
        "No optional init parameters, test defaults",
        dict(
            url='https://InvalidHostName',
            creds=None,
            default_namespace='root/cimv2',
            x509=None,
            ca_certs=None,
            no_verification=False,
            timeout=5,
            use_pull_operations=False,
            stats_enabled=False,
            proxies=None,
        ),
        ConnectionError, None
    ),
]


class TestConnectionError(object):  # pylint: disable=too-few-public-methods
    """
    Test construction of WBEMConnection and error exception when connection
    is executed with a function that communicates with the server.
    Only returns the exception message determined by the system since we have
    not control over these exceptions.
    """

    @pytest.mark.parametrize(
        'desc, init_kwargs, exp_exc, exp_exc_regex',
        TESTCASES_WBEMCONNECTION_ERROR)
    def test_init(self, desc, init_kwargs, exp_exc, exp_exc_regex):
        # pylint: disable=no-self-use,unused-argument
        """
        Test  WBEMConnection object Exceptions caused by Server connection
        failure.
        """
        exc = None
        try:
            assert exp_exc is not None
            with pytest.raises(exp_exc) as exc_info:

                # The code to be tested
                conn = WBEMConnection(**init_kwargs)
                _ = conn.EnumerateClasses()

            exc = exc_info.value
            exc_msg = str(exc)
            if exp_exc_regex:
                assert re.search(exp_exc_regex, exc_msg)
        finally:
            assert exc is not None



# -*- coding: utf-8 -*-
#
# (C) Copyright 2022 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
Pywbem mock tests for associations that crosses namespaces.  These
tests are separate because the multi-namespace capability was added in pywbem
1.1 so it was simpler to add a new test file. The tests
provide a test of the logic of processing these associations.
"""
from __future__ import absolute_import, print_function

import pytest

from ...utils import skip_if_moftab_regenerated
from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from pywbem._utils import _format  # noqa: E402

from ...utils import import_installed
pywbem = import_installed('pywbem')

from pywbem import DEFAULT_NAMESPACE, CIMInstanceName, CIMProperty, \
    CIMInstance, CIMError, CIM_ERR_NOT_FOUND, MOFRepositoryError, \
    Uint32  # noqa: E402

pywbem_mock = import_installed('pywbem_mock')
from pywbem_mock import FakedWBEMConnection  # noqa:E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# List of initially existing namespaces in the CIM repository
INITIAL_NAMESPACES = [DEFAULT_NAMESPACE]

# Literal form {"blah: 0} faster than dict(blah=0) but same functionality
# pylint: disable=use-dict-literal

#
# Class mof for associations and endpoints. This is created in all namespaces.
#
TST_CLASSES_MOF = """
    Qualifier Association : boolean = false,
        Scope(association),
        Flavor(DisableOverride, ToSubclass);

    Qualifier Description : string = null,
        Scope(any),
        Flavor(EnableOverride, ToSubclass, Translatable);

    Qualifier Key : boolean = false,
        Scope(property, reference),
        Flavor(DisableOverride, ToSubclass);

    class TST_Person{
            [Key, Description ("This is key prop")]
        string name;
        string extraProperty = "defaultvalue";
    };

    class TST_Personsub : TST_Person{
        string secondProperty = "empty";
        uint32 counter;
    };

    [ Description("Third endpoint for TST_MemberOfGroup3EP") ]
    class TST_THIRD_EP {
            [Key, Description ("This is key prop and endpoint name")]
        string name;
    };

    [ Description("Collection of Persons into a group") ]
    class TST_Group {
            [Key, Description ("This is key prop and group name")]
        string name;
    };

    [Association, Description("Association defines the relationship "
        "between people and groups with InstanceID as key.") ]
    class TST_MemberOfGroup {
        [key] string InstanceID;
        TST_Person ref member;
        TST_Group ref group;
            [Description("Extra parameter to allow ModifyInstance test")]
        string ExtraPropStr;
        uint32 ExtraPropInt;
    };

    [Association, Description(" Association defines the relationship "
        "between people and groups with reference properties as keys.") ]
    class TST_MemberOfGroupNoID {
        [key] TST_Person ref member;
        [key] TST_Group ref group;
            [Description("Extra parameter to allow ModifyInstance test")]
        string ExtraPropStr;
        uint32 ExtraPropInt;
    };

    [Association, Description(" Association defines the 3-way relationship "
        "between people, groups, and 3ep with reference properties as keys.") ]
    class TST_MemberOfGroup3EP {
        [key] string InstanceID;
        TST_Person ref member;
        TST_Group ref group;
        TST_THIRD_EP ref ep_3;
            [Description("Extra parameter to allow ModifyInstance test")]
        string ExtraPropStr;
        uint32 ExtraPropInt;

    };

"""

# MOF of Simple cross namespace association endpoint where association key is
# InstanceID. This defines the instances for one of the association endpoints,
# normally TST_Person
INSTANCE_INITIAL_ENDPOINTS = """
instance of TST_Person as $Mike { name = "Mike"; };

"""

# This defines the common endpoint and the association of
# TST_MemberOfGroup in ns1.
ASSOCIATION_MOF_SAME_NS = """
// association in same namespace

instance of TST_Group as $NS1G1
{
    name = "ns1group1";
};

instance of TST_MemberOfGroup as $G1MikeNS1
{
    InstanceID = "G1MikeNS1";
    member = $Mike;
    group = $NS1G1;
};

"""

# Association TST_MemberOfGroup in ns1 with person in ns1, group in ns2,
# and the association in ns1
ASSOCIATION_MOF_CROSS_NS = """
// association crosses namespaces

#pragma namespace ("ns2")
// target instance
instance of TST_Group as $NS2G1
{
    name = "ns2group1";
};

#pragma namespace ("ns1")
instance of TST_MemberOfGroup as $G1MikeNS2
{
    InstanceID = "G1MikeNS1";
    member = $Mike;
    group = $NS2G1;
    ExtraPropStr = "OrigValue";
    ExtraPropInt = 999;
};
"""

# Association TST_MemberOfGroup in ns1 with person in ns1, group in ns2,
# and the association in ns1 with extra properties in association. Used to
# test modify instance
ASSOCIATION_MOF_CROSS_NS_NO_EXTRAPROPS = """
// association crosses namespaces

#pragma namespace ("ns2")
// target instance
instance of TST_Group as $NS2G1
{
    name = "ns2group1";
};

#pragma namespace ("ns1")
instance of TST_MemberOfGroup as $G1MikeNS2
{
    InstanceID = "G1MikeNS1";
    member = $Mike;
    group = $NS2G1;

};
"""

# Association TST_MemberOfGroup in ns1 with person in ns1 and group in ns2
ASSOCIATION_MOF_CROSS_NS_CAUSES_CLASSERROR = """
// association crosses namespaces

#pragma namespace ("ns2")
// target instance
instance of TST_Group as $NS2G1
{
    name = "ns2group1";
};

#pragma namespace ("ns1")
instance of TST_MemberOfGroup as $G1MikeNS2
{
    InstanceID = "G1MikeNS1";
    member = $Mike;
    group = $NS2G1;
};
"""

# Association TST_MemberOfGroupNoID in ns1 with person in ns1 and group in ns2
# Assoc key is ref properties
ASSOCNOID_MOF_CROSS_NS = """
// association crosses namespaces

#pragma namespace ("ns2")
// target instance
instance of TST_Group as $NS2G1
{
    name = "ns2group1";
};

#pragma namespace ("ns1")
instance of TST_MemberOfGroupNoID as $G1MikeNS2A2
{
    member = $Mike;
    group = $NS2G1;
};

"""


# Association 3-way with TST_THIRD_EP endpoint and instance of
# TST_MemberOfGroup3EP
ASSOCIATIONX_3WAY_MOF_CROSS_NS = """
// association crosses namespaces

#pragma namespace ("ns2")
// target instance
instance of TST_Group as $NS2G1
{
    name = "ns2group1";
};

#pragma namespace ("ns3")
// target instance  of class TST_THIRD_EP
instance of TST_THIRD_EP as $NS3X1
{
    name = "ns3X1";
};

#pragma namespace ("ns1")
instance of TST_MemberOfGroup3EP as $G1MikeAxNs3
{
    InstanceID = "G1MikeAxNs3";
    member = $Mike;
    group = $NS2G1;
    ep_3 = $NS3X1;
};

"""


# Instance paths for instances of MenberOfGroupNoID, the association that
# uses the references as keys.  Too long to insert into tests themselves.
COMMON_MEMBEROFGROUPNOID = 'TST_MemberOfGroupNoID.group="/ns2:TST_Group.' \
    'name=\\"ns2group1\\"",member="/ns1:TST_Person.name=\\"Mike\\""'

MEMBEROFGROUPNOIDNS1 = '/ns1:' + COMMON_MEMBEROFGROUPNOID
MEMBEROFGROUPNOIDNS2 = '/ns2:' + COMMON_MEMBEROFGROUPNOID

# test variables to allow selectively executing tests.
OK = True
RUN = True
FAIL = False
VERBOSE = False


def create_repository(conn, namespaces, class_mof, inst_mof, assoc_inst_mof):
    """
    Create the namespaces, qualdecls, classes, instances for the test
    """
    # Create remaining namespaces
    for ns in namespaces:
        if ns not in conn.namespaces:
            conn.add_namespace(ns)

    skip_if_moftab_regenerated()

    # Compile the class mof in all namespaces
    for ns in namespaces:
        for mof in class_mof:
            conn.compile_mof_string(mof, ns)

    # Compile all instance mof together to utilize compiler alias capability
    # each of the components is a list.
    all_inst_mof = "\n".join(inst_mof) + "\n".join(assoc_inst_mof)

    conn.compile_mof_string(all_inst_mof)


def val_act_exp_paths(act_rtns, exp_rtn, ns, operation, source_cln):
    """
    Test that the  actual instance paths are exactly the same as
    the expected returns in exp_rtn[ns][operation][source_cln].

    Ignores host parameter of returned instance paths
    """

    assert isinstance(act_rtns, list)

    # Get expected returns for this ns, operation, and source_cln
    # from exp_rtn dictionary
    exp_rtns = exp_rtn[ns][operation][source_cln]

    # Convert URIs in exp_rtns to CIMInstanceName objects
    exp_objs = [CIMInstanceName.from_wbem_uri(p) for p in exp_rtns]

    assert len(act_rtns) == len(exp_rtns), \
        _format("ns:{}, operation:{}, source_cln:{}", ns, operation, source_cln)

    for instname in act_rtns:
        instname.host = None

    assert set(act_rtns) == set(exp_objs), \
        _format("ns:{}, operation:{}, tst_class:{}", ns, operation, source_cln)


def validate_instance_enumeration(conn, namespaces, enum_clns, exp_rtn):
    """
    Validate that EnumerateInstances returns correct instances. Note we only
    test instance names. The enum_clns is the list of all classnames for which
    EnumerateInstanceNames is to be executed.
    """
    for ns in namespaces:
        # Test EnumerateInstanceNames for all classes in enumeration
        for tst_cln in enum_clns:
            inst_paths = conn.EnumerateInstanceNames(tst_cln, namespace=ns)
            val_act_exp_paths(inst_paths, exp_rtn, ns, 'ENUM', tst_cln)


def validate_instance_associations(conn, namespaces, source_clns, exp_rtn):
    """
    Test that AssociationNames and ReferenceNames return correct instance
    paths
    """
    for ns in namespaces:
        # Test References and Associators for each classname in source_clns
        # This tests associations in both directions.
        for source_cln in source_clns:
            # Test only if there is an instance of source_cln
            if conn.EnumerateInstances(source_cln):
                source_paths = conn.EnumerateInstanceNames(source_cln,
                                                           namespace=ns)

                for source_path in source_paths:
                    # Compare ReferenceNames and AssociatonNames requests
                    inst_paths = conn.ReferenceNames(source_path)
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'REF',
                                      source_cln)
                    inst_paths = conn.AssociatorNames(source_path)
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'ASSOC',
                                      source_cln)

                    # Compare paths of References returns
                    insts = conn.References(source_path)
                    # Validate path and reference property same values
                    for inst in insts:
                        keys = inst.path.keybindings
                        for prop in inst.properties.values():
                            if prop.type == 'references':
                                if prop.name in keys:
                                    assert prop.value == keys[prop.name]
                    inst_paths = [i.path for i in insts]
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'REF',
                                      source_cln)

                    # Compare paths of Associator returns
                    insts = conn.Associators(source_path)
                    inst_paths = [r.path for r in insts]
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'ASSOC',
                                      source_cln)


# pylint: disable=bad-continuation


TESTCASES_MULTI_NAMESPACES = [
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * namespaces - List of namespaces in mock environment
    #   * class_mof - list of strings containing qualdecl and class mof to
    #     be compiled
    #   * inst_mof - list of strings containing instance mof to be compiled
    #   * assoc_inst_mof - list of strings  containing instances defining
    #     associations to be compiled. These strings must define the namespace
    #     for components that are not in the default namespace.
    #   * source classnames for ENUM, REF, ASSOC tests - The classnames that
    #     are the source classes for these operations
    #   * assoc_clns - The classname of the reference class being tested
    #   * delete_ns - If not None, defnes namespace from which association
    #     is deleted in delete test.  If None, deleted from creation ns.
    #   * exp_rtn: dictionary containing expected return objects (primarily
    #     CIMInstanceNames). The dictionary is defined as
    #     {<namespace>: <operation>: <source_class>: list of results instance
    #     names.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger

    (
        "Test association in one namespace. Assures model/test works",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_SAME_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            delete_ns=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': ['/ns1:TST_Group.name="ns1group1"'],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"'],
                         'TST_Group':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Person': ['/ns1:TST_Group.name="ns1group1"']}, },
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': [],
                         'TST_MemberOfGroup': []},
                     'REF':
                        {'TST_Person': []},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': []}, },
            },
        ),
        None, None, OK
    ),

    (
        "Test association cross namespaces with MemberOfGroup",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            delete_ns=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),

    (
        "Test association cross namespaces with MemberOfGroup delete from "
        "alternate namespace",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            delete_ns='ns2',
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),

    (
        "Test association TST_MemberOfGroupNoID (refs as keys) cross ns",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCNOID_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroupNoID',
            delete_ns=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroupNoID': [MEMBEROFGROUPNOIDNS1]},
                     'REF':
                        {'TST_Person': [MEMBEROFGROUPNOIDNS1]},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroupNoID': [MEMBEROFGROUPNOIDNS2]},
                     'REF':
                        {'TST_Group': [MEMBEROFGROUPNOIDNS2]},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),

    (
        "Test 3way association TST_MemberOfGroup3EP cross  3 namespaces",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2', 'ns3'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATIONX_3WAY_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group', 'TST_THIRD_EP'],
            assoc_clns='TST_MemberOfGroup3EP',
            delete_ns='ns1',
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_THIRD_EP': [],
                         'TST_MemberOfGroup3EP':
                             ['/ns1:TST_MemberOfGroup3EP.InstanceID='
                              '"G1MikeAxNs3"']},
                     'REF':
                        {'TST_Person':
                            ['/ns1:TST_MemberOfGroup3EP.InstanceID='
                             '"G1MikeAxNs3"']},
                     'ASSOC':
                        {'TST_Person': ['/ns2:TST_Group.name="ns2group1"',
                                        '/ns3:TST_THIRD_EP.name="ns3X1"'],
                         'TST_Group': [],
                         'TST_THIRD_EP': []}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         "TST_THIRD_EP": [],
                         'TST_MemberOfGroup3EP':
                             ['/ns2:TST_MemberOfGroup3EP.InstanceID='
                              '"G1MikeAxNs3"']},
                     'REF':
                        {'TST_Group':
                            ['/ns1:TST_MemberOfGroup3EP.InstanceID='
                             '"G1MikeAxNs3"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"',
                                       '/ns3:TST_THIRD_EP.name="ns3X1"'],
                         'TST_THIRD_EP': []}},
                'ns3':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': [],
                         'TST_THIRD_EP': ['/ns3:TST_THIRD_EP.name="ns3X1"'],
                         'TST_MemberOfGroup3EP':
                             ['/ns3:TST_MemberOfGroup3EP.InstanceID='
                              '"G1MikeAxNs3"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup3EP.InstanceID='
                             '"G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': [],
                         'TST_THIRD_EP': ['/ns1:TST_Person.name="Mike"',
                                          '/ns2:TST_Group.name="ns2group1"']}},
            },
        ),
        None, None, OK
    ),

    # TODO: Needs the following tests.
    #  Test with other instances and more associations in the test
    #  Lower priority:
    #    Test 3 way association.

]
# pylint: enable=bad-continuation

# Disabled because it is unhappy with indentation of multilevel dictionary
# lower levels.  Note that this occurs with minimum version tests but not with
# latest version tests


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_MULTI_NAMESPACES)
@simplified_test_function
def test_multiple_namespace_assoc(testcase, namespaces, class_mof, inst_mof,
                                  assoc_inst_mof, source_clns, assoc_clns,
                                  delete_ns, exp_rtn):
    # pylint: disable=unused-argument
    """
    Test the creation and deletion of instances of association that cross
    namespaces.
    """
    conn = FakedWBEMConnection(default_namespace=namespaces[0])

    # Create the repository from test components.  Creates qualdecls,
    # classes, endpoint instances, association instances in multiple
    # namespaces
    create_repository(conn, namespaces, class_mof, inst_mof, assoc_inst_mof)

    assert len(namespaces) == len(exp_rtn)

    # Create list of all classes for enumeration test. Note: This
    # creates a new list.  list() forces create of new object
    enum_clns = list(source_clns)
    enum_clns.append(assoc_clns)

    # Test that the creation of instances was correct.
    validate_instance_enumeration(conn, namespaces, enum_clns, exp_rtn)

    validate_instance_associations(conn, namespaces, source_clns, exp_rtn)

    # Test delete of association instances

    # delete the instance names of the associations in namespace defined by
    # delete_ns.  This should delete each of these associations in each
    # namespace
    deleted_assocs = []
    if delete_ns:
        ns = delete_ns
        inst_names = conn.EnumerateInstanceNames(assoc_clns, namespace=ns)
        for inst_name in inst_names:
            _ = conn.GetInstance(inst_name)
            conn.DeleteInstance(inst_name)
            deleted_assocs.append(inst_name)

        # Confirm inst_name deleted in all namespaces
        for namespace in namespaces:
            for assoc_name in deleted_assocs:
                try:
                    assoc_name.namespace = namespace
                    conn.GetInstance(assoc_name)
                    assert False, _format(
                        "Instance {} should not exist in ns {} after delete",
                        assoc_name, namespace)
                except CIMError as ce:
                    assert ce.status_code == CIM_ERR_NOT_FOUND


# pylint: disable=bad-continuation

TESTCASES_MULTI_NAMESPACES_ERRORS = [
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * namespaces - List of namespaces in mock environment
    #   * class_mof - list of strings containing qualdecl and class mof to
    #     be compiled
    #   * classes to remove from namespaces (dict <ns>: [<class names>]) before
    #     attempting to compile instances.
    #   * inst_mof - list of strings containing instance mof to be compiled
    #   * assoc_inst_mof
    #   * assoc_inst_obj - Instance definition in python. Used to create
    #     invalid instances
    #   * src_clns -
    #   * assoc_clns -
    #   * delete_ns -
    #   * exp_rtn -
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger

    (
        "Test fails with assoc class missing from second namespace",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            remove_classes={'ns2': ['TST_MemberOfGroup']},
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            assoc_inst_obj=None,
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroupNoID',
            delete_ns=None,
            exp_rtn=None
        ),
        MOFRepositoryError, None, OK
    ),

    (
        "Test fails,assoc instance already exists",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            remove_classes={'ns2': ['TST_MemberOfGroup']},
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            assoc_inst_obj=None,
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroupNoID',
            delete_ns=None,
            exp_rtn=None
        ),
        MOFRepositoryError, None, OK
    ),

    (
        "Test  bad end-points paths (do not exist) fails",
        dict(
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            remove_classes={'ns2': ['TST_MemberOfGroup']},
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[],
            # Association instance with  invalid reference properties
            assoc_inst_obj=CIMInstance(
                'TST_MemberOfGroup',
                path=CIMInstanceName(
                    "TST_MemberOfGroup", keybindings=dict(InstanceID="blah"),
                    namespace='ns1'),
                properties=dict(
                    InstanceID=CIMProperty("InstanceID", "blah"),
                    member=CIMProperty(
                        "member",
                        CIMInstanceName('TST_PERSON',
                                        keybindings=dict(name="fred"),
                                        namespace='ns1')),
                    group=CIMProperty(
                        'group',
                        CIMInstanceName('TST_Group',
                                        keybindings=dict(name="john"),
                                        namespace='ns2')))),
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            delete_ns=None,
            exp_rtn=None
        ),
        CIMError, None, OK
    ),

    (
        "Test  bad 2nd end-point path (instance does not exist) fails",
        dict(
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            remove_classes={'ns2': ['TST_MemberOfGroup']},
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[],
            # Association instance with  invalid reference properties
            assoc_inst_obj=CIMInstance(
                'TST_MemberOfGroup',
                path=CIMInstanceName(
                    "TST_MemberOfGroup", keybindings=dict(InstanceID="blah"),
                    namespace='ns1'),
                properties=dict(
                    InstanceID=CIMProperty("InstanceID", "blah"),
                    member=CIMProperty(
                        "member",
                        CIMInstanceName('TST_PERSON',
                                        keybindings=dict(name="Mike"),
                                        namespace='ns1')),
                    group=CIMProperty(
                        'group',
                        CIMInstanceName('TST_Group',
                                        keybindings=dict(name="john"),
                                        namespace='ns2')))),
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            delete_ns=None,
            exp_rtn=None
        ),
        CIMError, None, OK
    ),

    (
        "Test bad 2nd end-point path  namespace does not exist) fails",
        dict(
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            remove_classes={'ns2': ['TST_MemberOfGroup']},
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[],
            # Association instance with  invalid reference properties
            assoc_inst_obj=CIMInstance(
                'TST_MemberOfGroup',
                path=CIMInstanceName(
                    "TST_MemberOfGroup", keybindings=dict(InstanceID="blah"),
                    namespace='ns1'),
                properties=dict(
                    InstanceID=CIMProperty("InstanceID", "blah"),
                    member=CIMProperty(
                        "member",
                        CIMInstanceName('TST_PERSON',
                                        keybindings=dict(name="Mike"),
                                        namespace='ns1')),
                    group=CIMProperty(
                        'group',
                        CIMInstanceName('TST_Group',
                                        keybindings=dict(name="ns2group1"),
                                        namespace='InvalidNS')))),
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            delete_ns=None,
            exp_rtn=None
        ),
        CIMError, None, OK
    ),


    #  Test fail mode, Create assoc instance where instance already exists.
    #  Test fail mode where on create assoc instance exists already in
    #  second namespace

]
# pylint: enable=bad-continuation


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_MULTI_NAMESPACES_ERRORS)
@simplified_test_function
def test_multiple_namespace_assoc_errs(testcase, namespaces, class_mof,
                                       remove_classes, inst_mof,
                                       assoc_inst_mof, assoc_inst_obj,
                                       source_clns, assoc_clns,
                                       delete_ns, exp_rtn):

    # pylint: disable=unused-argument
    """
    Test the creation and deletion of instances of association that cross
    namespaces.
    """
    conn = FakedWBEMConnection(default_namespace=namespaces[0])

    # Create namespaces
    for ns in namespaces:
        if ns not in conn.namespaces:
            conn.add_namespace(ns)

    skip_if_moftab_regenerated()

    # Compile the class mof in all namespaces
    for ns in namespaces:
        for mof in class_mof:
            conn.compile_mof_string(mof, ns)

    # Delete classes defined by remove_classes to set up compile failures
    for ns, clns in remove_classes.items():
        for cln in clns:
            conn.DeleteClass(cln, ns)

    # Compile all instance mof together to utilize compiler alias capability
    # each of the components is a list.
    all_inst_mof = "\n".join(inst_mof) + "\n".join(assoc_inst_mof)

    conn.compile_mof_string(all_inst_mof, verbose=VERBOSE)

    # Create list of all classes for enumeration test. Note: This
    # creates a new list.
    enum_clns = list(source_clns)
    enum_clns.append(assoc_clns)

    # Create instance defined by assoc_inst_obj.
    if assoc_inst_obj:
        conn.CreateInstance(assoc_inst_obj)

    # Test that the creation of instances was correct.
    for ns in namespaces:
        # Test EnumerateInstanceNames for all classes in enumeration
        for tst_cln in enum_clns:
            inst_paths = conn.EnumerateInstanceNames(tst_cln, namespace=ns)
            val_act_exp_paths(inst_paths, exp_rtn, ns, 'ENUM', tst_cln)

        # Test References and Associators for each classname in source_clns
        for source_cln in source_clns:
            # Test only if there is an instance of source_cln
            if conn.EnumerateInstances(source_cln):
                source_paths = conn.EnumerateInstanceNames(source_cln,
                                                           namespace=ns)

                for source_path in source_paths:
                    # Compare ..Names requests
                    inst_paths = conn.ReferenceNames(source_path)
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'REF',
                                      source_cln)
                    inst_paths = conn.AssociatorNames(source_path)
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'ASSOC',
                                      source_cln)

                    # Compare paths of References returns
                    insts = conn.References(source_path)
                    # Validate path and reference property same values
                    for inst in insts:
                        keys = inst.path.keybindings
                        for prop in inst.properties.values():
                            if prop.type == 'references':
                                if prop.name in keys:
                                    assert prop.value == keys[prop.name]
                    inst_paths = [i.path for i in insts]
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'REF',
                                      source_cln)

                    # Compare paths of Associator returns
                    insts = conn.Associators(source_path)
                    inst_paths = [r.path for r in insts]
                    val_act_exp_paths(inst_paths, exp_rtn, ns, 'ASSOC',
                                      source_cln)


# pylint: disable=bad-continuation

TESTCASES_MULTI_NAMESPACES_MODIFY_INSTANCE = [
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * namespaces - List of namespaces in mock environment
    #   * class_mof - list of strings containing qualdecl and class mof to
    #     be compiled
    #   * inst_mof - list of strings containing instance mof to be compiled.
    #     These are the endpoint instances to be installed in ns1
    #   * assoc_inst_mof - list of strings  containing instances defining
    #     associations and other endpoints to be compiled. These strings must
    #     define the namespace for components that are not in the ns1 namespace.
    #   * source classnames for ENUM, REF, ASSOC tests - The classnames that
    #     are the source classes for these operations
    #   * assoc_clns - The classname of the reference class being tested
    #   * delete_ns - If not None, defnes namespace from which association
    #     is deleted in delete test.  If None, deleted from creation ns.
    #   * property_mods - If None defines a set of modifications to properties
    #     of an instance defined in a tuple where item 0 is the path to the
    #     instance to modify and item 1 is is a tuple with either:
    #     * one or more tuples where each tuple defines the property name and
    #       new value for the modification.
    #     * a CIMProperty that defines the name, type and value of the
    #       replacement property value.
    #     Prop_list to be passed to ModifyInstance. None, modify all properties
    #     defined as different (see property_mods), list of strings, modify
    #     only properties in the list, empty list, modify no properties.
    #   * del_assoc_instance_test: defines instance path of one assoc instance
    #     to delete (one of the shadowed instances) to test modify when
    #     shadow insts not complete.
    #   * exp_rtn: dictionary containing expected return objects (primarily
    #     CIMInstanceNames). The dictionary is defined as
    #     {<namespace>: <operation>: <source_class>: list of results instance
    #     names.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger

    (
        "Test association Modify in one namespace. Modify ExtraPropStr prop",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_SAME_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=('/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                           (("ExtraPropStr", "NewStringVal"),)),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': ['/ns1:TST_Group.name="ns1group1"'],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"'],
                         'TST_Group':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Person': ['/ns1:TST_Group.name="ns1group1"']}, },
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': [],
                         'TST_MemberOfGroup': []},
                     'REF':
                        {'TST_Person': []},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': []}, },
            },
        ),
        None, None, OK
    ),

    (
        "Test association modify cross namespaces with MemberOfGroup. No prop "
        "values in instance",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS_NO_EXTRAPROPS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=('/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                           (("ExtraPropStr", "NewStringVal"),
                            ("ExtraPropInt", Uint32(88888)))),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),

    (
        "Test association modify cross namespaces with MemberOfGroup modifiy 2 "
        "properties",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=('/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                           (("ExtraPropStr", "NewStringVal"),
                            ("ExtraPropInt", Uint32(88888)))),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),

    (
        "Test association modify cross namespaces with MemberOfGroup mod 2 prop"
        " and property list with both properties",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=('/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                           (("ExtraPropStr", "NewStringVal"),
                            ("ExtraPropInt", Uint32(88888)))),
            prop_list=["ExtraPropStr", "ExtraPropInt"],
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),


    (
        "Test association modify cross namespaces with MemberOfGroup delete "
        "from alternate namespace",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=('/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                           (("ExtraPropStr", "NewStringVal"),)),
            prop_list=None,
            delete_ns='ns2',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),

    (
        "Test association modify TST_MemberOfGroupNoID (refs as keys) cross ns",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCNOID_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroupNoID',
            property_mods=(MEMBEROFGROUPNOIDNS1,
                           (("ExtraPropStr", "NewStringVal"),)),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroupNoID':
                             [MEMBEROFGROUPNOIDNS1]},
                     'REF':
                        {'TST_Person':
                             [MEMBEROFGROUPNOIDNS1]},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroupNoID':
                             [MEMBEROFGROUPNOIDNS2]},
                     'REF':
                        {'TST_Group':
                            [MEMBEROFGROUPNOIDNS2]},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        None, None, OK
    ),

    (
        "Test 3way association modify TST_MemberOfGroup3EP cross  3 namespaces",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2', 'ns3'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATIONX_3WAY_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group', 'TST_THIRD_EP'],
            assoc_clns='TST_MemberOfGroup3EP',
            property_mods=('/ns1:TST_MemberOfGroup3EP.InstanceID="G1MikeAxNs3"',
                           (("ExtraPropStr", "NewStringVal"),)),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_THIRD_EP': [],
                         'TST_MemberOfGroup3EP':
                             ['/ns1:TST_MemberOfGroup3EP.InstanceID='
                              '"G1MikeAxNs3"']},
                     'REF':
                        {'TST_Person':
                            ['/ns1:TST_MemberOfGroup3EP.InstanceID='
                             '"G1MikeAxNs3"']},
                     'ASSOC':
                        {'TST_Person': ['/ns2:TST_Group.name="ns2group1"',
                                        '/ns3:TST_THIRD_EP.name="ns3X1"'],
                         'TST_Group': [],
                         'TST_THIRD_EP': []}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         "TST_THIRD_EP": [],
                         'TST_MemberOfGroup3EP':
                             ['/ns2:TST_MemberOfGroup3EP.InstanceID='
                              '"G1MikeAxNs3"']},
                     'REF':
                        {'TST_Group':
                            ['/ns1:TST_MemberOfGroup3EP.InstanceID='
                             '"G1MikeAxNs3"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"',
                                       '/ns3:TST_THIRD_EP.name="ns3X1"'],
                         'TST_THIRD_EP': []}},
                'ns3':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': [],
                         'TST_THIRD_EP': ['/ns3:TST_THIRD_EP.name="ns3X1"'],
                         'TST_MemberOfGroup3EP':
                             ['/ns3:TST_MemberOfGroup3EP.InstanceID='
                              '"G1MikeAxNs3"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup3EP.InstanceID='
                             '"G1MikeAxNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': [],
                         'TST_THIRD_EP': ['/ns1:TST_Person.name="Mike"',
                                          '/ns2:TST_Group.name="ns2group1"']}},
            },
        ),
        None, None, OK
    ),

    (
        "Test association modify cross namespaces with MemberOfGroup and"
        "ref prop modification",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=(
                '/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                # Modify group reference property
                (("group", CIMInstanceName("TST_Group", namespace='ns2',
                                           keybindings={"name":
                                                        "NotExist"})), )),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        CIMError, None, OK
    ),

    (
        "Test association modify cross namespaces with a shadow assoc inst "
        "missing",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=('/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                           ((CIMProperty("member", type="reference",
                                         value=None),))),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        CIMError, None, OK
    ),

    (
        "Test association modify cross namespaces with MemberOfGroup mod "
        "ref to Null",
        dict(
            # first namespace becomes the default namespace
            namespaces=['ns1', 'ns2'],
            class_mof=[TST_CLASSES_MOF],
            inst_mof=[INSTANCE_INITIAL_ENDPOINTS],
            assoc_inst_mof=[ASSOCIATION_MOF_CROSS_NS],
            source_clns=['TST_Person', 'TST_Group'],
            assoc_clns='TST_MemberOfGroup',
            property_mods=('/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"',
                           ((CIMProperty("member", type="reference",
                                         value=None),))),
            prop_list=None,
            delete_ns='ns1',
            delete_assoc_inst_test=None,
            exp_rtn={
                'ns1':
                    {'ENUM':
                        {'TST_Person': ['/ns1:TST_Person.name="Mike"'],
                         'TST_Group': [],
                         'TST_MemberOfGroup':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Person':
                             ['/ns1:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Group': [],
                         'TST_Person': ['/ns2:TST_Group.name="ns2group1"']}},
                'ns2':
                    {'ENUM':
                        {'TST_Person': [],
                         'TST_Group': ['/ns2:TST_Group.name="ns2group1"'],
                         'TST_MemberOfGroup':
                             ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'REF':
                        {'TST_Group':
                            ['/ns2:TST_MemberOfGroup.InstanceID="G1MikeNS1"']},
                     'ASSOC':
                        {'TST_Person': [],
                         'TST_Group': ['/ns1:TST_Person.name="Mike"']}}
            },
        ),
        CIMError, None, RUN
    ),


    # TODO: Needs the following tests.


]
# pylint: enable=bad-continuation

# Disabled this pylint because it is unhappy with indentation of multilevel
# dictionary lower levels.  Note that this occurs with minimum version tests
# but not with latest version tests


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_MULTI_NAMESPACES_MODIFY_INSTANCE)
@simplified_test_function
def test_multiple_namespace_assoc_modify(testcase, namespaces, class_mof,
                                         inst_mof, assoc_inst_mof, source_clns,
                                         assoc_clns, property_mods, prop_list,
                                         delete_ns, delete_assoc_inst_test,
                                         exp_rtn):
    # pylint: disable=unused-argument
    """
    Test the ModifyInstance of multi-namespace association instances defined
    in pywbem_mock.
    """
    conn = FakedWBEMConnection(default_namespace=namespaces[0])

    # Create the repository from test components.  Creates qualdecls,
    # classes, endpoint instances, association instances in multiple
    # namespaces
    create_repository(conn, namespaces, class_mof, inst_mof, assoc_inst_mof)

    assert len(namespaces) == len(exp_rtn)

    # Create list of all classes for enumeration test. Note: This
    # creates a new list.  list() forces create of new object
    enum_clns = list(source_clns)
    enum_clns.append(assoc_clns)

    # Test that the creation of instances was correct.
    validate_instance_enumeration(conn, namespaces, enum_clns, exp_rtn)

    validate_instance_associations(conn, namespaces, source_clns, exp_rtn)

    # if delete_assoc_inst_test exists delete the instance defined for the
    # test from the repository.  The modify should then fail.
    if delete_assoc_inst_test:
        ns = delete_assoc_inst_test.namespace
        instance_store = conn.cimrepository.get_instance_store(ns)
        instance_store.delete(delete_assoc_inst_test)

    # Modify the association instance defined. Test by getting it from repo and
    # comparing values
    if property_mods:
        inst_name = CIMInstanceName.from_wbem_uri(property_mods[0])
        modified_inst = conn.GetInstance(inst_name)
        mods = property_mods[1]
        for pmod in mods:
            if isinstance(pmod, tuple):
                modified_inst[pmod[0]] = pmod[1]
            elif isinstance(pmod, CIMProperty):
                modified_inst[pmod.name] = pmod
            else:
                assert False, "Invalid property_mod test parameter"

        if prop_list is None:
            conn.ModifyInstance(modified_inst)
        else:
            conn.ModifyInstance(modified_inst, PropertyList=prop_list)

        modified_inst_rtn = conn.GetInstance(modified_inst.path)
        for prop in modified_inst:
            assert modified_inst_rtn[prop] == modified_inst[prop]

        for pmod in mods:
            if prop_list and pmod in prop_list:
                assert modified_inst_rtn[pmod[0]] == pmod[1]

    # validate enumeration and associations with modified instance
    validate_instance_enumeration(conn, namespaces, enum_clns, exp_rtn)

    validate_instance_associations(conn, namespaces, source_clns, exp_rtn)

    # Test delete of association instances

    # delete the instance names of the associations in namespace defined by
    # delete_ns.  This should delete each of these associations in each
    # namespace
    deleted_assocs = []
    if delete_ns:
        ns = delete_ns
        inst_names = conn.EnumerateInstanceNames(assoc_clns, namespace=ns)
        for inst_name in inst_names:
            _ = conn.GetInstance(inst_name)
            conn.DeleteInstance(inst_name)
            deleted_assocs.append(inst_name)

        # Confirm inst_name deleted in all namespaces
        for namespace in namespaces:
            for assoc_name in deleted_assocs:
                try:
                    assoc_name.namespace = namespace
                    conn.GetInstance(inst_name)
                    assert False, _format("Instance {} should not exist in "
                                          "ns {} after delete", inst_name,
                                          namespace)
                except CIMError as ce:
                    assert ce.status_code == CIM_ERR_NOT_FOUND


# pylint: disable=bad-continuation



"""
Utility functions for memory size of objects.
"""

import sys
import itertools
import struct
import collections

DEBUG = False  # Print debug messages in total_sizeof()


def _dict_iter(dict_):
    """
    Return flat iterable of dict keys and values.
    """
    return itertools.chain.from_iterable(dict_.items())


def total_sizeof(obj, iters=None):
    """
    Return the memory size in Bytes of an object and all of its attributes.
    For collections that do not expose their items as attributes, the size
    of the items is also accounted for.

    A number of built-in collection classes that do not expose their items as
    attributes are recognized automatically:

        tuple
        list
        collections.deque
        dict
        set
        frozenset

    Items returned by dictionary view classes are intentionally not accounted
    for, because they are accounted on the dictionary object.

    Additional collection classes can be specified with the 'iters' parameter.
    That is only needed for collection classes that are not derived from the
    automatically recognized collection classes and that do not expose their
    items as attributes. Note that this excludes most if not all user-defined
    collection classes written in pure Python.

    Parameters:

      obj (object): The object whose size is to be returned.

      iters (dict): Iteration functions for additional collections.
        Key is the collection class, value is a function or bound method that
        takes an object of the collection class (or subclass thereof) as a
        single argument, and iterates over its items.

        Example::

            iters = {
                SomeContainerClass: iter,
                OtherContainerClass: OtherContainerClass.get_elements,
            }

    Original idea from https://code.activestate.com/recipes/577504/
    """

    def _debug_enter():
        """
        Do things needed when entering sizeof().
        """
        if DEBUG:
            nesting[0] += 1

    def _debug_leave():
        """
        Do things needed when leaving sizeof().
        """
        if DEBUG:
            nesting[0] -= 1

    def _debug_msg(msg, obj, size):
        """
        Print debug message about the progress of determining the object size.
        """
        if DEBUG:
            o_str = repr(obj) if isinstance(obj, (str, bool, int)) else ''
            print("Debug: {}: {} {} object {} at {}: {}".
                  format(nesting[0], msg, type(obj), o_str, id(obj),
                         size if size is not None else ''))

    # Built-in collection types where the collection items do not show up in
    # attributes.
    all_iters = collections.OrderedDict([
        (tuple, iter),
        (list, iter),
        (dict, _dict_iter),
        (collections.deque, iter),
        (set, iter),
        (frozenset, iter),
    ])

    # Additionally provided iteration functions take precedence
    if iters:
        all_iters.update(iters)

    seen = set()  # Track which objects have already been seen

    default_size = sys.getsizeof(0)
    ref_size = struct.calcsize('P')  # pylint: disable=no-member

    if DEBUG:
        nesting = [0]

    def sizeof(o):
        # pylint: disable=invalid-name
        """
        Return total size of the object.
        """
        _debug_enter()

        if id(o) in seen:
            # Do not double count the same object, but account for referencing
            # the object
            s = ref_size
            _debug_msg("Size of reference to already seen", o, s)
            _debug_leave()
            return s
        seen.add(id(o))

        # Size of the object without its attributes or collection items
        s = sys.getsizeof(o, default_size)
        _debug_msg("Base size of", o, s)

        # Add size of collection items, recursively
        for cls, iterfunc in all_iters.items():
            if isinstance(o, cls):
                _debug_msg("Calculating size of collection items of", o, None)
                ds = sum(map(sizeof, iterfunc(o)))
                s += ds
                _debug_msg("Adding size of collection items of", o, ds)
                break

        # Add size of slotted attributes, recursively
        if getattr(o, '__slots__', None):
            _debug_msg("Calculating size of slotted attributes of", o, None)
            ds = sum(sizeof(getattr(o, attr_name))
                     for attr_name in o.__slots__ if hasattr(o, attr_name))
            s += ds
            _debug_msg("Adding size of slotted attributes of", o, ds)

        # Add size of non-slotted attributes, recursively
        if getattr(o, '__dict__', None):
            _debug_msg("Calculating size of non-slotted attributes of", o, None)
            ds = sizeof(o.__dict__)
            s += ds
            _debug_msg("Adding size of non-slotted attributes of", o, ds)

        _debug_leave()
        return s

    return sizeof(obj)



#!/usr/bin/env python

"""
Measure performance of sending indications to the pywbem listener.
"""

from __future__ import absolute_import

import sys
from time import sleep
import statistics
import requests
import pytest

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ..utils import import_installed, post_bsl
from ..elapsed_timer import ElapsedTimer
pywbem = import_installed('pywbem')
from pywbem import WBEMListener  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


def create_indication_data(msg_id, sequence_number, protocol_ver):
    """
    Create a test indication from the template and input attributes.
    """

    data_template = """<?xml version="1.0" encoding="utf-8" ?>
    <CIM CIMVERSION="2.0" DTDVERSION="2.4">
      <MESSAGE ID="{msg_id}" PROTOCOLVERSION="{protocol_ver}">
        <SIMPLEEXPREQ>
          <EXPMETHODCALL NAME="ExportIndication">
            <EXPPARAMVALUE NAME="NewIndication">
              <INSTANCE CLASSNAME="CIM_AlertIndication">
                <PROPERTY NAME="Severity" TYPE="string">
                  <VALUE>high</VALUE>
                </PROPERTY>
                <PROPERTY NAME="SequenceNumber" TYPE="string">
                  <VALUE>{sequence_number}</VALUE>
                </PROPERTY>
              </INSTANCE>
            </EXPPARAMVALUE>
          </EXPMETHODCALL>
        </SIMPLEEXPREQ>
      </MESSAGE>
    </CIM>"""

    return data_template.format(
        sequence_number=sequence_number,
        protocol_ver=protocol_ver, msg_id=msg_id)


# Global variables used to communicate between the test case function and
# the process_indication() function running in context of the listener
# thread. These must be global, because in Python 2, closure variables
# cannot be modified.
RCV_COUNT = 0
RCV_ERRORS = False


def process_indication(indication, host):
    # pylint: disable=unused-argument
    """
    This function gets called by the listener when an indication is
    received.

    It tests the received indication sequence number against the RCV_COUNT
    which should catch any duplicated or missing indication, since the
    counters would no longer match in such cases.

    This function is invoked in context of the listener thread. Therefore,
    it is defined as a local function in order to be able to use the
    variables from its outer function, specifically the RCV_COUNT and
    RCV_ERRORS variables.

    For the same reason, this function does not report assertion failures
    by means of raising exceptions, but by printing a message and setting
    RCV_ERRORS to an error message, so the test function can check it and
    raise assertion errors.
    """

    # Note: Global variables that are modified must be declared global
    global RCV_COUNT  # pylint: disable=global-statement
    global RCV_ERRORS  # pylint: disable=global-statement

    try:

        send_count = int(indication.properties['SequenceNumber'].value)
        if send_count != RCV_COUNT:
            print("Error in process_indication(): Assertion error: "
                  "Unexpected SequenceNumber in received indication: "
                  "got {}, expected {}".format(send_count, RCV_COUNT))
            sys.stdout.flush()
            RCV_ERRORS = True

        RCV_COUNT += 1

    except Exception as exc:  # pylint: disable=broad-except
        print("Error in process_indication(): {}: {}".
              format(exc.__class__.__name__, exc))
        sys.stdout.flush()
        RCV_ERRORS = True


def send_indications(host, http_port, send_count):
    """
    Send a number of indications and return the elapsed time for that.
    """

    # Note: Global variables that are modified must be declared global
    global RCV_COUNT  # pylint: disable=global-statement
    global RCV_ERRORS  # pylint: disable=global-statement

    url = 'http://{}:{}'.format(host, http_port)
    cim_protocol_version = '1.4'
    headers = {
        'Content-Type': 'application/xml; charset=utf-8',
        'CIMExport': 'MethodRequest',
        'CIMExportMethod': 'ExportIndication',
        'Accept-Encoding': 'Identity',
        'CIMProtocolVersion': cim_protocol_version,
    }

    timer = ElapsedTimer()

    RCV_COUNT = 0
    RCV_ERRORS = False

    for i in range(send_count):

        msg_id = i
        payload = create_indication_data(msg_id, i, cim_protocol_version)

        try:
            response = post_bsl(url, headers=headers, data=payload)
        except requests.exceptions.RequestException as exc:
            msg = ("Testcase sending indication #{} raised {}: {}".
                   format(i, exc.__class__.__name__, exc))
            new_exc = AssertionError(msg)
            new_exc.__cause__ = None  # Disable to see original traceback
            raise new_exc

        if response.status_code != 200:
            msg = ("Testcase sending indication #{} failed with HTTP "
                   "status {}".format(i, response.status_code))
            raise AssertionError(msg)

    endtime = timer.elapsed_sec()

    # Give the listener thread some time to process all indications
    sleep(0.5)

    assert not RCV_ERRORS, \
        "Errors occurred in process_indication(), as printed to stdout"

    assert send_count == RCV_COUNT, \
        "Mismatch between total send count {} and receive count {}". \
        format(send_count, RCV_COUNT)

    return endtime


@pytest.mark.parametrize(
    "send_count",
    [1, 2, 3, 4, 5, 10, 100, 200, 300, 400, 500]
)
def test_WBEMListener_send_indications(send_count):
    """
    Test performance of sending indications to the pywbem.WBEMListener.
    """

    if send_count > 100 and sys.platform == 'win32':
        pytest.skip("Skipping test due to lengthy elapsed time")

    host = 'localhost'
    http_port = 50000

    listener = WBEMListener(host, http_port)
    listener.add_callback(process_indication)
    listener.start()

    # Warm up
    send_indications(host, http_port, 10)

    repetitions = 5
    try:
        times = []
        for _ in range(0, repetitions):
            time = send_indications(host, http_port, send_count)
            times.append(time)
        mean_time = statistics.mean(times)
        stdev_time = statistics.stdev(times)
        mean_rate = send_count / mean_time

        print("\nSent {} indications ({} repetitions): mean: {:.3f} s, "
              "stdev: {:.3f} s, mean rate: {:.0f} ind/s".
              format(send_count, repetitions, mean_time, stdev_time, mean_rate))
        sys.stdout.flush()

    finally:
        listener.stop()

        # Give some time to free up the port
        sleep(1.0)



"""
Utility functions for measuring time.
"""

from datetime import datetime
try:
    import psutil
except ImportError:
    # No psutil support e.g. on CygWin
    psutil = None


class Timer(object):
    """
    Timer measuring elapsed time (= wall clock time) and CPU time
    (= execution time).

    Elapsed time is measured using 'datetime.datetime.now()'.
    Its resolution depends on the platform.

    CPU time is measured using the 'psutil' package using the 'cpu_times()'
    method on the current process. The CPU time includes user time and system
    time across all CPUs of the system in context of the current process.
    Its resolution depends on the platform.
    """

    def __init__(self):
        """
        Initialize the timer. It is initially not running.
        """
        self.start_datetime = None
        self.start_cpu_times = None
        self.stop_datetime = None
        self.stop_cpu_times = None
        self.process = psutil.Process() if psutil else None

    @property
    def running(self):
        """
        Return bool indicating whether the timer is running.
        """
        return self.start_datetime is not None and self.stop_datetime is None

    def start(self):
        """
        Start the timer.
        """
        self.start_datetime = datetime.now()
        if psutil:
            self.start_cpu_times = self.process.cpu_times()
        self.stop_datetime = None
        if psutil:
            self.stop_cpu_times = None

    def stop(self):
        """
        Stop the timer.

        The timer must be running.
        """
        stop_datetime = datetime.now()
        if psutil:
            stop_cpu_times = self.process.cpu_times()
        if not self.running:
            raise ValueError("Timer is not running")
        self.stop_datetime = stop_datetime
        if psutil:
            self.stop_cpu_times = stop_cpu_times

    def elapsed_ms(self):
        """
        Get the elapsed time in milliseconds between last start and stop.

        The timer must not be running.

        Returns:
           float: Elapsed time in milliseconds.
        """
        if self.running:
            raise ValueError("Timer is running")
        dt = self.stop_datetime - self.start_datetime
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000 + \
            dt.microseconds / 1000.0

    def cpu_ms(self):
        """
        Get the CPU time (= execution time) in milliseconds between last
        start and stop.

        The CPU time is the total spent on all CPUs of the system and
        includes user time plus system time.

        The timer must not be running.

        If the platform does not support the 'psutil' package, 0.0 is returned.

        Returns:
           float: CPU time in milliseconds.
        """
        if self.running:
            raise ValueError("Timer is running")
        if not psutil:
            return 0.0
        dt_user = self.stop_cpu_times.user - self.start_cpu_times.user
        dt_system = self.stop_cpu_times.system - self.start_cpu_times.system
        dt = dt_user + dt_system
        return dt * 1000



"""
Context manager that measures resource consumption of its body.
"""

import tracemalloc

from .timers import Timer


class ResourceMeasurement(object):
    """
    Context manager that measures resource consumption of its body.

    Example::

        with ResourceMeasurement() as rm:
            # code to be measured
        print(rm.cpu_time_ms)

    Attributes:

      peak_size (int): Peak size in Bytes used during execution of the body, as
        reported by tracemalloc.

      elapsed_time_ms (int): Elapsed (wall clock) time in ms for the execution
        of the body.

      cpu_time_ms (int): CPU (execution) time in ms for the execution
        of the body.
    """

    def __init__(self):
        self.timer = Timer()
        self.peak_size = None
        self.elapsed_time_ms = None
        self.cpu_time_ms = None

    def __enter__(self):
        self.timer.start()
        tracemalloc.start()  # pylint: disable=no-member  # On pypy3
        return self

    def __exit__(self, exc_type, exc_value, traceback):

        # pylint: disable=no-member  # On pypy3
        _, self.peak_size = tracemalloc.get_traced_memory()
        self.timer.stop()
        tracemalloc.stop()  # pylint: disable=no-member  # On pypy3
        self.elapsed_time_ms = self.timer.elapsed_ms()
        self.cpu_time_ms = self.timer.cpu_ms()
        return False  # re-raise any exceptions
